local ADDON, Package = ...

local Kernel = {}
local String = {}
local Enum = {}
local Utils = {}
local Observable = {
  observers = {}
}
local Widget = {}

--------------------------------------------------------------------------------
-- Kernel
--------------------------------------------------------------------------------
function Kernel.is_table(any) return type(any) == "table" end

function Kernel.is_number(arg) return type(arg) == "number" end

function Kernel.is_string(arg) return type(arg) == "string" end

function Kernel.is_boolean(arg) return type(arg) == "boolean" end

function Kernel.noop() end

function Kernel.dump(...)
  if not IsAddOnLoaded("Blizzard_DebugTools") then
    UIParentLoadAddOn("Blizzard_DebugTools")
  end

  for at = 1, select("#", ...) do
    DevTools_Dump(select(at, ...))
  end
end

--------------------------------------------------------------------------------
-- String
--------------------------------------------------------------------------------
function String.startcase(value)
  return string.gsub(" " .. value, "%W%l", string.upper):gsub("[_ ]", "")
end

--------------------------------------------------------------------------------
-- Enum
--------------------------------------------------------------------------------
function Enum.clone(enumerable)
  local stash = {}

  for k, v in pairs(enumerable) do
    if Kernel.is_table(v) then
      stash[k] = Enum.clone(v)
    else
      stash[k] = v
    end
  end

  return stash
end

function Enum.merge(to, from, override)
  local stash = Enum.clone(to)

  for k, v in pairs(from) do
    if Kernel.is_table(v) and Kernel.is_table(stash[k]) then
      stash[k] = Enum.merge(stash[k], v, override)
    else
      if stash[k] == nil or override ~= false then
        stash[k] = v
      end
    end
  end

  return stash
end

function Enum.omit(enumerable, keys, depth)
  local stash = {}
  local depth = depth or 0

  for k, v in pairs(enumerable) do
    if not Enum.is_member(keys, k) then
      if Kernel.is_table(v) and depth > 0 then
        stash[k] = Enum.omit(v, keys, depth - 1)
      else
        stash[k] = v
      end
    end
  end

  return stash
end

function Enum.rename(enumerable, from, to)
  local stash = {}

  for k, v in pairs(enumerable) do
    local i = k == from and to or k

    if Kernel.is_table(v) then
      stash[i] = Enum.rename(v, from, to)
    else
      stash[i] = v
    end
  end

  return stash
end

function Enum.filter(enumerable, value)
  local stash = {}

  for _, entry in pairs(enumerable) do
    if entry ~= value then
      table.insert(stash, entry)
    end
  end

  return stash
end

function Enum.is_member(enumerable, value)
  for _, entry in pairs(enumerable) do
    if entry == value then
      return true
    end
  end

  return false
end

function Enum.concat(left, right)
  local stash = Enum.clone(left)

  for _, entry in pairs(right) do
    table.insert(stash, entry)
  end

  return stash
end

--------------------------------------------------------------------------------
-- Utils
--------------------------------------------------------------------------------
function Utils.purge_keyboard_focus()
  local frame = GetCurrentKeyBoardFocus()
  if frame and frame.ClearFocus then
    frame:ClearFocus()
  end
end

--------------------------------------------------------------------------------
-- Observable
--------------------------------------------------------------------------------
function Observable.broadcast(scope, ...)
  if not Observable.observers[scope] then
    Observable.observers[scope] = {}
  end

  for at = 1, #Observable.observers[scope] do
    Observable.observers[scope][at](...)
  end
end

function Observable.subscribe(scope, observer)
  if not Observable.observers[scope] then
    Observable.observers[scope] = {}
  end

  if not Enum.is_member(Observable.observers[scope], observer) then
    table.insert(Observable.observers[scope], observer)
  end
end

function Observable.unsubscribe(scope, observer)
  if not Observable.observers[scope] then
    Observable.observers[scope] = {}
  end

  Observable.observers[scope] = Enum.filter(Observable.observers[scope], observer)
end

--------------------------------------------------------------------------------
-- Widget
--------------------------------------------------------------------------------
function Widget.create_window(optional_attrs, parent)
  local default_attrs = {
    text = "",
    size = {128, 128},
    name = nil
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", attrs.name, parent, "BackdropTemplate")
  frame:Hide()
  frame:EnableKeyboard(true)
  frame:EnableMouse(true)
  frame:EnableMouseWheel(true)
  frame:SetToplevel(true)
  frame:SetMovable(true)
  frame:SetFrameStrata("DIALOG")
  frame:SetSize(unpack(attrs.size))
  frame:SetScript("OnMouseUp", function (self) Utils.purge_keyboard_focus() end)

  local title = CreateFrame("Frame", nil, frame)
  title:EnableMouse(true)
  title:SetScript("OnDragStart", function (self) self:GetParent():StartMoving() end)
  title:SetScript("OnDragStop", function (self) self:GetParent():StopMovingOrSizing() end)
  title:RegisterForDrag("LeftButton")
  title:SetPoint("TOP", 0, 6)
  title:SetSize(128, 32)

  local title_background = frame:CreateTexture(nil, "ARTWORK")
  title_background:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
  title_background:SetPoint("TOP", 0, 12)
  title_background:SetSize(256, 64)

  local title_text = frame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
  title_text:SetPoint("TOP", title_background, 0, -14)
  title_text:SetText(attrs.text)

  local close_button_border = frame:CreateTexture(nil, "BORDER")
  close_button_border:SetTexture(131073)
  close_button_border:SetSize(32, 32)
  close_button_border:SetPoint("TOPRIGHT", -8, -8)

  local close_button = CreateFrame("Button", nil, frame, "UIPanelCloseButton")
  close_button:SetPoint("TOPRIGHT", -4, -4)

  frame.title = title
  frame.title_background = title_background
  frame.title_text = title_text
  frame.close_button_border = close_button_border
  frame.close_button = close_button

  return frame
end

function Widget.create_tab_button(optional_attrs, parent)
  local default_attrs = {
    text = ""
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Button", nil, parent)
  frame:SetSize(64, 32)

  local left = frame:CreateTexture(nil, "BACKGROUND")
  left:SetTexture(130942)
  left:SetTexCoord(0.0, 0.25, 0.0, 1.0)
  left:SetSize(16, 32)
  left:SetPoint("TOPLEFT")

  local middle = frame:CreateTexture(nil, "BACKGROUND")
  middle:SetTexture(130942)
  middle:SetTexCoord(0.25, 0.75, 0.0, 1.0)
  middle:SetSize(44, 32)
  middle:SetPoint("LEFT", left, "RIGHT")

  local right = frame:CreateTexture(nil, "BACKGROUND")
  right:SetTexture(130942)
  right:SetTexCoord(0.75, 1.0, 0.0, 1.0)
  right:SetSize(16, 32)
  right:SetPoint("LEFT", middle, "RIGHT")

  local highlight = frame:CreateTexture(nil, "BACKGROUND")
  highlight:SetTexture(136580)
  highlight:SetPoint("LEFT", left, 0, -7)
  highlight:SetPoint("RIGHT", right, 0, -7)
  highlight:SetBlendMode("ADD")

  local pushed = frame:CreateTexture(nil, "BACKGROUND")
  pushed:SetTexture(375515)
  pushed:SetPoint("LEFT", left, 0, -7)
  pushed:SetPoint("RIGHT", right, 0, -7)
  pushed:SetBlendMode("ADD")

  local font_string = frame:CreateFontString()
  font_string:SetPoint("LEFT", left, 8, -6)
  font_string:SetPoint("RIGHT", right, -8, -6)
  font_string:SetJustifyH("CENTER")

  frame:SetHighlightTexture(highlight)
  frame:SetPushedTexture(pushed)
  frame:SetPushedTextOffset(0, 0)
  frame:SetNormalFontObject("GameFontNormalSmall")
  frame:SetFontString(font_string)
  frame:SetText(attrs.text)

  return frame
end

function Widget.create_tab_panel(optional_attrs, parent)
  local default_attrs = {}
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", nil, parent)
  frame:Hide()

  return frame
end

function Widget.create_listbox(optional_attrs, parent)
  local default_attrs = {
    name = nil,
    margin = {top = 0, right = 0, left = 0, bottom = 0}
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Frame", attrs.name, parent, "BackdropTemplate")
  frame:SetPoint("TOPLEFT", attrs.margin.left, -attrs.margin.top)
  frame:SetPoint("BOTTOMRIGHT", -attrs.margin.right - 22, attrs.margin.bottom)

  local scrollable = CreateFrame("ScrollFrame", nil, frame, "UIPanelScrollFrameTemplate")
  scrollable:SetAllPoints(frame)

  local child = CreateFrame("Frame", nil, scrollable)
  child:SetSize(frame:GetWidth(), frame:GetHeight())

  local collection = CreateFrame("Frame", nil, scrollable)
  collection:SetAllPoints(scrollable)

  scrollable.Preserve_OnVerticalScroll = scrollable:GetScript("OnVerticalScroll")
  scrollable.Preserve_OnScrollRangeChanged = scrollable:GetScript("OnScrollRangeChanged")
  scrollable.Override_OnVerticalScroll = function (self, offset)
    self:Preserve_OnVerticalScroll(offset)
    self.items = {self.collection:GetChildren()}
    self.item_height = #self.items > 0 and math.floor(self.items[1]:GetHeight()) or 0
    self.items_offset = math.floor((offset / math.max(1, self.item_height)) + 0.5)
    self:OnUpdate(self.data, self.items, self.items_offset)
  end
  scrollable.Override_OnScrollRangeChanged = function (self, xrange, yrange)
    self:Preserve_OnScrollRangeChanged(xrange, math.max(0, (#self.data - #self.items) * self.item_height))
    self:GetScrollChild():SetHeight(#self.data * self.item_height)
    self.ScrollBar:SetObeyStepOnDrag(true)
    self.ScrollBar:SetValueStep(self.item_height)
    self.ScrollBar:SetStepsPerPage(#self.items - 1)
  end
  scrollable:SetScript("OnVerticalScroll", scrollable.Override_OnVerticalScroll)
  scrollable:SetScript("OnScrollRangeChanged", scrollable.Override_OnScrollRangeChanged)
  scrollable:SetScrollChild(child)

  frame.SetEvent = function(self, type, handler)
    if type == "OnUpdate" then
      self.scrollable.OnUpdate = handler
    end
  end
  frame.SetData = function (self, data)
    self.scrollable.data = data or {}
  end
  frame.Update = function (self)
    self.scrollable:Override_OnVerticalScroll(self.scrollable.ScrollBar:GetValue())
    self.scrollable:Override_OnScrollRangeChanged(0, 0)
  end

  frame.scrollable = scrollable
  frame.scrollable.child = child
  frame.scrollable.collection = collection

  return frame
end

function Widget.create_button(optional_attrs, parent)
  local default_attrs = {
    text = "",
    size = {80, 24},
    tooltip = nil
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("Button", nil, parent, "UIPanelButtonTemplate")
  frame:SetText(attrs.text)
  frame:SetSize(unpack(attrs.size))
  frame.tooltipText = attrs.tooltip

  return frame
end

function Widget.create_checkbox(optional_attrs, parent)
  local default_attrs = {
    text = "",
    checked = false
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})

  local frame = CreateFrame("CheckButton", nil, parent, "UICheckButtonTemplate")
  frame:SetChecked(attrs.checked)
  frame.text:SetText(attrs.text)

  return frame
end

function Widget.create_searchbox(optional_attrs, parent)
  local default_attrs = {
    name = nil
  }
  local attrs = Enum.merge(default_attrs, optional_attrs or {})
  local frame = CreateFrame("EditBox", attrs.name, parent, "SearchBoxTemplate")

  return frame
end

Package.Kernel = Kernel
Package.String = String
Package.Enum = Enum
Package.Utils = Utils
Package.Observable = Observable
Package.Widget = Widget
Package.L = Package[GetLocale()] or Package["enUS"]