local ADDON, Package = ...

local Module = {}
local Registry = {}

local L = Package.L
local clone = Package.Enum.clone
local concat = Package.Enum.concat
local create_tab_button = Package.Widget.create_tab_button
local create_tab_panel = Package.Widget.create_tab_panel
local create_button = Package.Widget.create_button
local create_checkbox = Package.Widget.create_checkbox

local DIALOGBOX_BACKDROP = {
  bgFile = "Interface/DialogFrame/UI-DialogBox-Background",
  edgeFile = "Interface/DialogFrame/UI-DialogBox-Border",
  tile = true,
  tileSize = 32,
  tileEdge = true,
  edgeSize = 32,
  insets = {top = 8, right = 8, left = 8, bottom = 8}
}

local TOOLTIP_BACKDROP = {
  bgFile = "Interface/Tooltips/UI-Tooltip-Background",
  edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
  tile = true,
  tileSize = 16,
  tileEdge = true,
  edgeSize = 16,
  insets = {top = 4, right = 4, left = 4, bottom = 4}
}

local RACE_ICONS = {
  HUMAN = {"raceicon-human-male", "raceicon-human-female"},
  DWARF = {"raceicon-dwarf-male", "raceicon-dwarf-female"},
  GNOME = {"raceicon-gnome-male", "raceicon-gnome-female"},
  NIGHTELF = {"raceicon-nightelf-male", "raceicon-nightelf-female"},
  BLOODELF = {"raceicon-bloodelf-male", "raceicon-bloodelf-female"},
  DRAENEI = {"raceicon-draenei-male", "raceicon-draenei-female"},
  TAUREN = {"raceicon-tauren-male", "raceicon-tauren-female"},
  SCOURGE = {"raceicon-undead-male", "raceicon-undead-female"},
  TROLL = {"raceicon-troll-male", "raceicon-troll-female"},
  ORC = {"raceicon-orc-male", "raceicon-orc-female"}
}

local is_string = Package.Kernel.is_string
local merge = Package.Enum.merge
local create_window = Package.Widget.create_window
local create_listbox = Package.Widget.create_listbox
local create_searchbox = Package.Widget.create_searchbox

local agent = CreateFrame("Frame")

local state = {
  history = {},
  initiator = "target",
  transaction = nil
}

local LA_PALMA_FONT_COLOR = CreateColor(0.1, 0.6, 0.1)

function Module.format_character_level(level)
  local color = GetQuestDifficultyColor(level)

  return CreateColor(color.r, color.g, color.b):WrapTextInColorCode(format("%02d", level))
end

function Module.format_character_class_and_name(class, name)
  return RAID_CLASS_COLORS[class]:WrapTextInColorCode(name)
end

function Module.format_character_guild(guild)
  if guild and guild[1] then
    return LA_PALMA_FONT_COLOR:WrapTextInColorCode("<" .. guild[1] .. ">")
  else
    return ""
  end
end

function Module.format_zone_and_subzone(zone, subzone)
  if subzone == zone or subzone == "" then
    return zone
  else
    return zone .. " - " .. subzone
  end
end

function Module.format_trade_money_text(amount)
  local stash = {}
  local g = floor(amount / (COPPER_PER_SILVER * SILVER_PER_GOLD))
  local s = floor((amount - (g * COPPER_PER_SILVER * SILVER_PER_GOLD)) / COPPER_PER_SILVER)
  local c = mod(amount, COPPER_PER_SILVER)

  if g > 0 then
    table.insert(stash, FormatLargeNumber(g) .. GOLD_AMOUNT_SYMBOL)
  end

  if s > 0 then
    table.insert(stash, s .. SILVER_AMOUNT_SYMBOL)
  end

  if c > 0 or (s == 0 and g == 0) then
    table.insert(stash, c .. COPPER_AMOUNT_SYMBOL)
  end

  return table.concat(stash, " ")
end

function Module.format_trade_item_text(name, quantity, enchantment)
  local stash = name

  if quantity > 1 then
    stash = stash .. "x" .. tostring(quantity)
  end

  if enchantment then
    stash = stash .. " @ " .. enchantment
  end

  return stash
end

function Module.get_character_icon_by_race_and_sex(race, sex)
  return RACE_ICONS[race][sex]
end

function Module.get_zone_pvp_color(pvp_type)
  if pvp_type == "sanctuary" then
    return 0.41, 0.8, 0.94
  elseif pvp_type == "arena" then
    return 1.0, 0.1, 0.1
  elseif pvp_type == "friendly" then
    return 0.1, 1.0, 0.1
  elseif pvp_type == "hostile" then
    return 1.0, 0.1, 0.1
  elseif pvp_type == "contested" then
    return 1.0, 0.7, 0.0
  else
    return NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b
  end
end

function Module.get_id_from_item_link(item_link)
  if item_link then
    return item_link:match("item:(%d+)")
  end
end

function Module.get_initiator()
  return state.initiator
end

function Module.get_timestamp()
  return GetServerTime()
end

function Module.get_character(unit_id)
  return {
    realm = GetRealmName(),
    guild = {GetGuildInfo(unit_id)},
    faction = UnitFactionGroup(unit_id):upper(),
    class = select(2, UnitClass(unit_id)):upper(),
    race = select(2, UnitRace(unit_id)):upper(),
    level = UnitLevel(unit_id),
    name = UnitName(unit_id),
    sex = UnitSex(unit_id) - 1
  }
end

function Module.get_character_snapshot(unit_id)
  return {
    player = Module.get_character("player"),
    target = Module.get_character("npc")
  }
end

function Module.get_map_snapshot(unit_id)
  local position = CreateVector2D(0, 0)
  local pvp_type = GetZonePVPInfo()
  local subzone = GetSubZoneText()
  local zone = GetZoneText() or UNKNOWN
  local id = C_Map.GetBestMapForUnit("player")

  if id then
    position = C_Map.GetPlayerMapPosition(id, "player")
  end

  return {
    zone = zone,
    subzone = subzone,
    pvp_type = pvp_type,
    position = {x = position.x, y = position.y}
  }
end

function Module.get_trade_item(slot_id, unit_id)
  local GetTradeItemInfo = ({player = GetTradePlayerItemInfo, target = GetTradeTargetItemInfo})[unit_id]
  local GetTradeItemLink = ({player = GetTradePlayerItemLink, target = GetTradeTargetItemLink})[unit_id]
  local name, _, quantity, _, enchantment_is_usable, can_lose_transmog_or_enchantment = GetTradeItemInfo(slot_id)
  local enchantment = ({player = enchantment_is_usable, target = can_lose_transmog_or_enchantment})[unit_id]
  local link = GetTradeItemLink(slot_id)
  local id = Module.get_id_from_item_link(link)

  if id and name and quantity then
    return {
      id = id,
      name = name,
      quantity = quantity,
      enchantment = is_string(enchantment) and enchantment or nil
    }
  else
    return {}
  end
end

function Module.get_trade_money(unit_id)
  local GetTradeMoney = ({player = GetPlayerTradeMoney, target = GetTargetTradeMoney})[unit_id]

  return GetTradeMoney()
end

function Module.get_trade_snapshot()
  local stash = {
    player = {money = 0, items = {}},
    target = {money = 0, items = {}}
  }

  stash.player.money = Module.get_trade_money("player")
  stash.target.money = Module.get_trade_money("target")

  for at = 1, MAX_TRADE_ITEMS do
    table.insert(stash.player.items, Module.get_trade_item(at, "player"))
    table.insert(stash.target.items, Module.get_trade_item(at, "target"))
  end

  return stash
end

function Module.transform_transaction_to_payload(unit_id)
  local payload = {}
  local name = state.transaction.character[unit_id].name
  local money = state.transaction.trade[unit_id].money
  local items = state.transaction.trade[unit_id].items

  local unique_items = {}
  local unique_items_index = {}

  for slot_id = 1, MAX_TRADE_ITEMS do
    local item = clone(items[slot_id])

    if item and item.name and item.quantity then
      if not unique_items_index[item.name] then
        table.insert(unique_items, item)

        unique_items_index[item.name] = #unique_items
      else
        local unique_item = unique_items[unique_items_index[item.name]]

        unique_item.quantity = unique_item.quantity + item.quantity
        unique_item.enchantment = unique_item.enchantment or item.enchantment
      end
    end
  end

  for slot_id = 1, MAX_TRADE_ITEMS do
    local item = unique_items[slot_id]

    if item and item.name and item.quantity then
      table.insert(payload, {key = "trade-item", value = Module.format_trade_item_text(item.name, item.quantity, item.enchantment)})
    end
  end

  if money > 0 then
    table.insert(payload, 1, {key = "trade-money", value = Module.format_trade_money_text(money)})
  end

  if #payload > 0 then
    table.insert(payload, 1, {key = "character-name", value = name})
  end

  return payload
end

function Module.accept_transaction()
  table.insert(state.history, 1, state.transaction)

  if #state.history > 512 then
    table.remove(state.history)
  end
end

function Module.update_transaction()
  state.transaction = {
    map = Module.get_map_snapshot(),
    trade = Module.get_trade_snapshot(),
    character = Module.get_character_snapshot(),
    timestamp = Module.get_timestamp(),
    initiator = Module.get_initiator()
  }
end

function Module.cancel_transaction()
  state.initiator = "target"
  state.transaction = nil
end

function Module.notify_accept_transaction()
  local player = Module.transform_transaction_to_payload("player")
  local target = Module.transform_transaction_to_payload("target")

  local payload = concat(player, target)

  local mailbox = {}
  local message = ""

  for at = 1, #payload do
    local entry = payload[at]

    if entry.key == "character-name" then
      entry.value = "<" .. entry.value .. "> "
    elseif entry.key == "trade-money" then
      entry.value = "" .. entry.value .. " "
    elseif entry.key == "trade-item" then
      entry.value = "[" .. entry.value .. "] "
    end

    local ahead = message .. entry.value

    if ahead:len() > 251 then -- 4 bytes reserved
      table.insert(mailbox, message:trim())

      message = entry.value
    else
      message = ahead
    end
  end

  if message:len() > 0 then
    table.insert(mailbox, message:trim())
  end

  if #mailbox > 1 then
    for at = 1, #mailbox do
      SendChatMessage(format("%d/%d %s", at, #mailbox, mailbox[at]), "WHISPER", nil, state.transaction.character.target.name)
    end
  elseif #mailbox > 0 then
    SendChatMessage(table.remove(mailbox), "WHISPER", nil, state.transaction.character.target.name)
  end
end

function Module.notify_cancel_transaction()
  SendChatMessage(ERR_TRADE_CANCELLED, "WHISPER", nil, state.transaction.character.target.name)
end

function Module.create_tabs(parent)
  local tabs = {}
  local tabs_attrs = {
    {name = "history", text = L["history"]},
    {name = "options", text = L["options"]}
  }

  for at = 1, #tabs_attrs do
    local tab = {}

    tab.name = tabs_attrs[at].name

    tab.button = create_tab_button({text = tabs_attrs[at].text}, parent)
    tab.button:SetID(at)
    tab.button:SetPoint("TOPLEFT", (at - 1) * 74, 30)

    tab.panel = create_tab_panel(nil, parent)
    tab.panel:SetID(at)
    tab.panel:SetAllPoints(parent)

    tabs[at] = tab
  end

  for at = 1, #tabs do
    tabs[at].button:SetScript("OnClick", function (self)
      local id = self:GetID()

      for at = 1, #tabs do
        tabs[at].button:SetButtonState("NORMAL", false)
        tabs[at].panel:Hide()

        tabs[id].button:SetButtonState("PUSHED", true)
        tabs[id].panel:Show()
      end
    end)
  end

  return tabs
end

function Module.create_history_listbox_item(parent)
  local frame = CreateFrame("Button", nil, parent)
  frame:SetPoint("LEFT", 0, 0)
  frame:SetPoint("RIGHT", 0, 0)
  frame:SetHeight(60)

  local background = frame:CreateTexture(nil, "BACKGROUND")
  background:SetTexture("Interface\\FrameGeneral\\UI-Background-Rock")
  background:SetPoint("TOPLEFT", 0, -12)
  background:SetPoint("BOTTOMRIGHT", 0, 0)
  background:SetVertexColor(0.6, 0.6, 0.6)

  local id = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  id:SetSize(36, 12)
  id:SetPoint("TOPLEFT", 0, 0)
  id:SetJustifyH("LEFT")
  id:SetTextColor(0, 1, 1)

  local timestamp = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  timestamp:SetSize(80, 12)
  timestamp:SetPoint("TOPLEFT", id, "TOPRIGHT", 4, 0)
  timestamp:SetJustifyH("LEFT")
  timestamp:SetTextColor(0.486, 0.518, 0.541)

  local map_title = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  map_title:SetSize(264, 12)
  map_title:SetPoint("TOPLEFT", timestamp, "TOPRIGHT", 4, 0)
  map_title:SetJustifyH("RIGHT")

  local map_position = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  map_position:SetSize(80, 12)
  map_position:SetPoint("TOPLEFT", map_title, "TOPRIGHT", 4, 0)
  map_position:SetJustifyH("RIGHT")
  map_position:SetTextColor(0.486, 0.518, 0.541)

  local character_player_icon = frame:CreateTexture(nil, "ARTWORK")
  character_player_icon:SetSize(48, 48)
  character_player_icon:SetPoint("BOTTOMLEFT", 0, 0)
  character_player_icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

  local character_target_icon = frame:CreateTexture(nil, "ARTWORK")
  character_target_icon:SetSize(48, 48)
  character_target_icon:SetPoint("BOTTOMLEFT", 236, 0)
  character_target_icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

  local character_player_meta = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  character_player_meta:SetSize(180, 12)
  character_player_meta:SetPoint("BOTTOMLEFT", character_player_icon, "BOTTOMRIGHT", 4, 36)
  character_player_meta:SetJustifyH("LEFT")

  local character_target_meta = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  character_target_meta:SetSize(180, 12)
  character_target_meta:SetPoint("BOTTOMLEFT", character_target_icon, "BOTTOMRIGHT", 4, 36)
  character_target_meta:SetJustifyH("LEFT")

  local trade_player_money = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  trade_player_money:SetSize(180, 12)
  trade_player_money:SetPoint("BOTTOMLEFT", character_player_icon, "BOTTOMRIGHT", 4, 24)
  trade_player_money:SetJustifyH("LEFT")

  local trade_target_money = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny")
  trade_target_money:SetSize(180, 12)
  trade_target_money:SetPoint("BOTTOMLEFT", character_target_icon, "BOTTOMRIGHT", 4, 24)
  trade_target_money:SetJustifyH("LEFT")

  local trade_player_items = {}
  local relative_to = character_player_icon
  local padding_left = 4
  for at = 1, MAX_TRADE_ITEMS do
    local item = CreateFrame("Button", nil, frame)
    item:SetSize(24, 24)
    item:SetPoint("BOTTOMLEFT", relative_to, "BOTTOMRIGHT", padding_left, 0)
    item:SetHighlightTexture("Interface\\Buttons\\UI-Quickslot2")
    item:GetHighlightTexture():SetTexCoord(0.2, 0.8, 0.2, 0.8)
    item:SetScript("OnMouseUp", function (self, button)
      if self.data and self.data.id and button == "LeftButton" then
        local name, link = GetItemInfo(self.data.id)

        if name and link then
          HandleModifiedItemClick(link)
        end
      end
    end)
    item:SetScript("OnEnter", function (self)
      if self.data and self.data.id then
        local name, link = GetItemInfo(self.data.id)

        if name and link then
          GameTooltip:SetOwner(self, "ANCHOR_LEFT")
          GameTooltip:SetHyperlink(link)

          if self.data.enchantment then
            GameTooltip:AddLine(ITEM_PROPOSED_ENCHANT:format(self.data.enchantment), 0.1, 1, 0.1)
          end

          GameTooltip:Show()
        end
      end
    end)
    item:SetScript("OnLeave", function (self)
      GameTooltip:Hide()
    end)

    local quantity = item:CreateFontString(nil, "ARTWORK", "NumberFont_Shadow_Tiny")
    quantity:SetSize(24, 24)
    quantity:SetText("3")
    quantity:SetPoint("TOPLEFT", item, -2, 1)
    quantity:SetJustifyH("RIGHT")
    quantity:SetJustifyV("BOTTOM")
    quantity:SetTextColor(1, 1, 1)
    quantity:Hide()

    relative_to = item
    padding_left = 2

    item.quantity = quantity

    table.insert(trade_player_items, item)
  end

  local trade_target_items = {}
  local relative_to = character_target_icon
  local padding_left = 4
  for at = 1, MAX_TRADE_ITEMS do
    local item = CreateFrame("Button", nil, frame)
    item:SetSize(24, 24)
    item:SetPoint("BOTTOMLEFT", relative_to, "BOTTOMRIGHT", padding_left, 0)
    item:SetHighlightTexture("Interface\\Buttons\\UI-Quickslot2")
    item:GetHighlightTexture():SetTexCoord(0.2, 0.8, 0.2, 0.8)
    item:SetScript("OnMouseUp", function (self, button)
      if self.data and self.data.id and button == "LeftButton" then
        local name, link = GetItemInfo(self.data.id)

        if name and link then
          HandleModifiedItemClick(link)
        end
      end
    end)
    item:SetScript("OnEnter", function (self)
      if self.data and self.data.id then
        local name, link = GetItemInfo(self.data.id)

        if name and link then
          GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
          GameTooltip:SetHyperlink(link)

          if self.data.enchantment then
            GameTooltip:AddLine(ITEM_PROPOSED_ENCHANT:format(self.data.enchantment), 0.1, 1, 0.1)
          end

          GameTooltip:Show()
        end
      end
    end)
    item:SetScript("OnLeave", function (self)
      GameTooltip:Hide()
    end)

    local quantity = item:CreateFontString(nil, "ARTWORK", "NumberFont_Shadow_Tiny")
    quantity:SetSize(24, 24)
    quantity:SetText("3")
    quantity:SetPoint("TOPLEFT", item, -2, 1)
    quantity:SetJustifyH("RIGHT")
    quantity:SetJustifyV("BOTTOM")
    quantity:SetTextColor(1, 1, 1)
    quantity:Hide()

    relative_to = item
    padding_left = 2

    item.quantity = quantity

    table.insert(trade_target_items, item)
  end

  local trade_player_enchant_slot = trade_player_items[TRADE_ENCHANT_SLOT]:CreateTexture(nil, "ARTWORK", nil, 7)
  trade_player_enchant_slot:SetAllPoints(trade_player_items[TRADE_ENCHANT_SLOT])
  trade_player_enchant_slot:SetVertexColor(0.1, 1.0, 0.1)
  trade_player_enchant_slot:SetTexCoord(0.2, 0.8, 0.2, 0.8)
  trade_player_enchant_slot:SetTexture("Interface\\Buttons\\UI-Quickslot2")

  local trade_target_enchant_slot = trade_target_items[TRADE_ENCHANT_SLOT]:CreateTexture(nil, "ARTWORK", nil, 7)
  trade_target_enchant_slot:SetAllPoints(trade_target_items[TRADE_ENCHANT_SLOT])
  trade_target_enchant_slot:SetVertexColor(0.1, 1.0, 0.1)
  trade_target_enchant_slot:SetTexCoord(0.2, 0.8, 0.2, 0.8)
  trade_target_enchant_slot:SetTexture("Interface\\Buttons\\UI-Quickslot2")

  frame.id = id
  frame.timestamp = timestamp
  frame.map_title = map_title
  frame.map_position = map_position

  frame.character_player_icon = character_player_icon
  frame.character_player_meta = character_player_meta
  frame.character_target_icon = character_target_icon
  frame.character_target_meta = character_target_meta

  frame.trade_player_money = trade_player_money
  frame.trade_target_money = trade_target_money
  frame.trade_player_items = trade_player_items
  frame.trade_target_items = trade_target_items

  frame.slot = "Interface\\Buttons\\UI-Quickslot"

  return frame
end

function Module.on_history_listbox_update(self, data, items, items_offset)
  for at, line in pairs(items) do
    local cursor = at + items_offset
    local index = #data - cursor + 1
    local entry = data[cursor]

    if cursor <= #data then
      line.id:SetFormattedText("%03d", index)
      line.timestamp:SetText(date("%m-%d %H:%M", entry.timestamp))
      line.map_title:SetText(Module.format_zone_and_subzone(entry.map.zone, entry.map.subzone))
      line.map_title:SetTextColor(unpack({Module.get_zone_pvp_color(entry.map.pvp_type)}))
      line.map_position:SetFormattedText("%05.2f•%05.2f", entry.map.position.x * 100, entry.map.position.y * 100)
      line.character_player_icon:SetAtlas(Module.get_character_icon_by_race_and_sex(entry.character.player.race, entry.character.player.sex), false, "NEAREST")
      line.character_player_icon:SetDesaturated(entry.initiator == "target")
      line.character_player_meta:SetFormattedText("%s %s %s",
        Module.format_character_level(entry.character.player.level),
        Module.format_character_class_and_name(entry.character.player.class, entry.character.player.name),
        Module.format_character_guild(entry.character.player.guild)
      )

      line.character_target_icon:SetAtlas(Module.get_character_icon_by_race_and_sex(entry.character.target.race, entry.character.target.sex), false, "NEAREST")
      line.character_target_icon:SetDesaturated(entry.initiator == "player")
      line.character_target_meta:SetFormattedText("%s %s %s",
        Module.format_character_level(entry.character.target.level),
        Module.format_character_class_and_name(entry.character.target.class, entry.character.target.name),
        Module.format_character_guild(entry.character.target.guild)
      )

      line.trade_player_money:SetText(GetMoneyString(entry.trade.player.money, true))
      line.trade_target_money:SetText(GetMoneyString(entry.trade.target.money, true))

      for at = 1, MAX_TRADE_ITEMS do
        local item_icon = GetItemIcon(entry.trade.player.items[at].id)

        if item_icon then
          line.trade_player_items[at]:SetNormalTexture(item_icon)
          line.trade_player_items[at]:GetNormalTexture():SetTexCoord(0, 1, 0, 1)
          line.trade_player_items[at].data = entry.trade.player.items[at]
        else
          line.trade_player_items[at]:SetNormalTexture(line.slot)
          line.trade_player_items[at]:GetNormalTexture():SetTexCoord(0.2, 0.8, 0.2, 0.8)
          line.trade_player_items[at].data = nil
        end

        if (entry.trade.player.items[at].quantity or 0) > 1 then
          line.trade_player_items[at].quantity:SetText(entry.trade.player.items[at].quantity)
          line.trade_player_items[at].quantity:Show()
        else
          line.trade_player_items[at].quantity:Hide()
        end

        local item_icon = GetItemIcon(entry.trade.target.items[at].id)

        if item_icon then
          line.trade_target_items[at]:SetNormalTexture(item_icon)
          line.trade_target_items[at]:GetNormalTexture():SetTexCoord(0, 1, 0, 1)
          line.trade_target_items[at].data = entry.trade.target.items[at]
        else
          line.trade_target_items[at]:SetNormalTexture(line.slot)
          line.trade_target_items[at]:GetNormalTexture():SetTexCoord(0.2, 0.8, 0.2, 0.8)
          line.trade_target_items[at].data = nil
        end

        if (entry.trade.target.items[at].quantity or 0) > 1 then
          line.trade_target_items[at].quantity:SetText(entry.trade.target.items[at].quantity)
          line.trade_target_items[at].quantity:Show()
        else
          line.trade_target_items[at].quantity:Hide()
        end
      end

      line:Show()
    else
      line:Hide()
    end
  end
end

function Module.initialize()
  local storage = MedusaTradeStorage

  local window = create_window({text = ADDON, size = {534, 424}, name = ADDON .. "History"}, UIParent)
  window:SetPoint("CENTER")
  window:SetClampedToScreen(false)
  window:SetBackdrop(DIALOGBOX_BACKDROP)

  local panel = CreateFrame("Frame", nil, window, "BackdropTemplate")
  panel:SetBackdrop(TOOLTIP_BACKDROP)
  panel:SetBackdropColor(0.1, 0.1, 0.1, 0.5)
  panel:SetBackdropBorderColor(0.5, 0.5, 0.5)
  panel:SetPoint("TOPLEFT", 12, -56)
  panel:SetPoint("BOTTOMRIGHT", -12, 12)

  local tabs = Module.create_tabs(panel)

  local history_listbox = create_listbox({margin = {top = 32, right = 8, left = 8, bottom = 8}, name = ADDON .. "HistoryListBox"}, tabs[1].panel)
  history_listbox:SetEvent("OnUpdate", Module.on_history_listbox_update)
  history_listbox:SetData(state.history)

  for at = 1, 5 do
    local item = Module.create_history_listbox_item(history_listbox.scrollable.collection)
    item:SetPoint("TOP", 0, -(item:GetHeight() + 4) * (at - 1))
  end

  history_listbox:Update()
  history_listbox:Show()

  local history_searchbox = create_searchbox({name = ADDON .. "HistorySearchBox"}, tabs[1].panel)
  history_searchbox:SetPoint("TOP", 0, -8)
  history_searchbox:SetSize(160, 18)
  history_searchbox:HookScript("OnTextChanged", function (self) Module.update() end)

  local use_transaction_notification_checkbox = create_checkbox({text = L["use_transaction_notification"]}, tabs[2].panel)
  use_transaction_notification_checkbox:SetScript("OnClick", function (self) storage["use_transaction_notification"] = self:GetChecked() end)
  use_transaction_notification_checkbox:SetScript("OnShow", function (self) self:SetChecked(storage["use_transaction_notification"]) end)
  use_transaction_notification_checkbox:SetPoint("TOPLEFT", 8, -8)
  use_transaction_notification_checkbox.text:SetWidth(tabs[2].panel:GetWidth() - 56)
  use_transaction_notification_checkbox.text:SetJustifyH("LEFT")

  -- select tab
  tabs[1].button:GetScript("OnClick")(tabs[1].button)

  table.insert(UISpecialFrames, window:GetName())
end

function Module.update()
  local search_term = MedusaTradeHistorySearchBox:GetText():trim()
  local search_results = {}

  for _, entry in pairs(state.history) do
    local map_zone = entry.map.zone
    local map_subzone = entry.map.subzone
    local character_target_name = entry.character.target.name
    local character_target_guild = entry.character.target.guild and entry.character.target.guild[1] or ""

    if character_target_name:find(search_term) or character_target_guild:find(search_term) then
      table.insert(search_results, entry)
    elseif map_zone:find(search_term) or map_subzone:find(search_term) then
      table.insert(search_results, entry)
    else
      for slot_id = 1, MAX_TRADE_ITEMS do
        local player_item = entry.trade.player.items[slot_id]
        local target_item = entry.trade.target.items[slot_id]
        local player_item_name = player_item and player_item.name or ""
        local target_item_name = target_item and target_item.name or ""
        local player_item_enchantment = player_item and player_item.enchantment or ""
        local target_item_enchantment = target_item and target_item.enchantment or ""

        if player_item_name:find(search_term) or target_item_name:find(search_term)
        or player_item_enchantment:find(search_term) or target_item_enchantment:find(search_term) then
          table.insert(search_results, entry)
          break
        end
      end
    end
  end

  if #search_results > 0 or search_term:len() > 0 then
    MedusaTradeHistoryListBox:SetData(search_results)
  else
    MedusaTradeHistoryListBox:SetData(state.history)
  end

  MedusaTradeHistoryListBox:Update()
end

function Module.load_storage()
  local default_schema = {
    version = 2.0,
    use_transaction_notification = true
  }

  local default_schema_per_character = {
    version = 2.0,
    history = {}
  }

  MedusaTradeStorage = merge(MedusaTradeStorage or {}, default_schema, false)
  MedusaTradeStoragePerCharacter = merge(MedusaTradeStoragePerCharacter or {}, default_schema_per_character, false)

  state.history = MedusaTradeStoragePerCharacter.history
end

function Module.run()
  agent:SetScript("OnEvent", Module.on_event)
  agent:RegisterEvent("PLAYER_LOGIN")
  agent:RegisterEvent("ADDON_LOADED")
  agent:RegisterEvent("UI_INFO_MESSAGE")
  agent:RegisterEvent("TRADE_SHOW")
  agent:RegisterEvent("TRADE_PLAYER_ITEM_CHANGED")
  agent:RegisterEvent("TRADE_TARGET_ITEM_CHANGED")
  agent:RegisterEvent("TRADE_ACCEPT_UPDATE")
  agent:RegisterEvent("TRADE_UPDATE")

  hooksecurefunc("InitiateTrade", function ()
    state.initiator = "player"
  end)
end

function Module.on_event(self, event, ...)
  Registry[event](self, ...)
end

function Registry.PLAYER_LOGIN()
  Module.initialize()
end

function Registry.ADDON_LOADED(self, ...)
  if ADDON == ... then
    Module.load_storage()

    local addon = WrapTextInColorCode(..., "ff91d7f2")
    local version = WrapTextInColorCode("v" .. GetAddOnMetadata(..., "Version"), "ff91d7f2")
    local slashes = WrapTextInColorCode(SLASH_MedusaTrade1 .. " " .. SLASH_MedusaTrade2, "ff72f2e5")

    DEFAULT_CHAT_FRAME:AddMessage(addon .. " " .. version .. " " .. slashes )
  end
end

function Registry.UI_INFO_MESSAGE(self, error_type, message)
  local storage = MedusaTradeStorage

  -- CANCEL
  if error_type == 226 then
    if storage["use_transaction_notification"] then
      Module.notify_cancel_transaction()
    end

    Module.cancel_transaction()
  end

  -- ACCEPT
  if error_type == 227 then
    if storage["use_transaction_notification"] then
      Module.notify_accept_transaction()
    end

    Module.accept_transaction()
    Module.cancel_transaction()
    Module.update()
  end
end

function Registry.TRADE_SHOW()
  Module.update_transaction()
end

function Registry.TRADE_PLAYER_ITEM_CHANGED()
  Module.update_transaction()
end

function Registry.TRADE_TARGET_ITEM_CHANGED()
  Module.update_transaction()
end

function Registry.TRADE_ACCEPT_UPDATE()
  Module.update_transaction()
end

function Registry.TRADE_UPDATE()
  Module.update_transaction()
end

Module.run()

SLASH_MedusaTrade1 = "/medusa_trade"

if GetLocale() == "koKR" then
  SLASH_MedusaTrade2 = "/거래내역"
else
  SLASH_MedusaTrade2 = ""
end

SlashCmdList[ADDON] = function ()
  MedusaTradeHistory:SetShown(not MedusaTradeHistory:IsShown())
end
